#!/usr/bin/env python

"""
Some examples of how the code in cards.py works.
"""

__author__ = "Christopher Potts"
__date__ = "2011-09-11"
__copyright__ = "Copyright 2011, Christopher Potts"
__credits__ = "Thanks to Karl Schultz for designing the data collecting program, and \
               to David Clausen, Alex Djalali, Sven Lauer, Victoria Schwanda, Chriz Czyzewicz, \
               and the rest of the SUBTLE team for their help with data collection. \
               This research was supported in part by ONR grant No. N00014-10-1-0109 and \
               ARO grant No. W911NF-07-1-0216."
__license__ = "Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License: \
               http://creativecommons.org/licenses/by-nc-sa/3.0/"
__version__ = "1.0"
__maintainer__ = "Christopher Potts"
__email__ = "See the author's website"

######################################################################

import re
import numpy
from collections import defaultdict
from operator import itemgetter
from cards import *

######################################################################

SRC_DIRNAME = 'transcripts'

def stats():
    """
    Gathers a lot of basic statistical information about the
    corpus. Output is to standard output.
    """    
    event_types = defaultdict(int)
    utterance_lengths = []
    game_lengths = []
    vocab = defaultdict(int)
    corpus = Corpus(SRC_DIRNAME)
    for trans in corpus.iter_transcripts():
        game_lengths.append(len(trans.events))        
        for event in trans.iter_events():
            event_types[event.action] += 1
            if event.action == UTTERANCE:
                words = event.parse_contents()
                utterance_lengths.append(len(words))
                for word in words:
                    vocab[word] += 1
    # Output:
    print 'Event types:'
    for key, val in sorted(event_types.items()):
        print key, val
    # Game length:
    print 'Mean game length:', numpy.mean(game_lengths)
    print 'Median game length:', numpy.median(game_lengths)
    print 'Std for game length:', numpy.std(game_lengths)
    # Utterances:
    print 'Mean utterance length:', numpy.mean(utterance_lengths)
    print 'Median utterance length:', numpy.median(utterance_lengths)
    print 'Std for utterance length:', numpy.std(utterance_lengths)
    # Vocab:
    print 'Vocab size:', len(vocab)
    print 'Word count:', sum(vocab.values())
    
######################################################################

CARD_RE = re.compile(r"(%s)(%s)" % (Tokenizer.ranks, Tokenizer.suits))

def parse_card(card):
    """Utility for parsing cards into rank and suit."""    
    rank, suit = CARD_RE.search(card).groups()
    return (rank, suit)

######################################################################

def all_attested_hands_to_csv(output_filename):
    """
    Creates a CSV file containing information about the hands that
    players hold.
    """
    # Output dictionary mapping card tuples to counts:
    counts = defaultdict(int)
    corpus = Corpus(SRC_DIRNAME)
    # Iterate through at the transcript level:
    for trans in corpus.iter_transcripts():
        ## This will map player names to holdings:
        holdings = defaultdict(list)
        for event in trans.iter_events():        
            if event.action == PICKUP:
                # PICKUP events are (x,y,card) triples with parse_contents():
                card = event.parse_contents()[2]
                holdings[event.agent].append(card)
                current_holdings = tuple(sorted(list(set(holdings[event.agent]))))
                counts[current_holdings] += 1
            elif event.action == DROP:
                card = event.parse_contents()[2]
                try:
                    holdings[event.agent].remove(card)
                except:
                    print "In %s, %s dropped %s but holdings are %s" % (trans.filename, event.agent, card, holdings[event.agent])
    # To CSV:
    csvwriter = csv.writer(file(output_filename, 'w'))
    # Header:
    csvwriter.writerow(['Hand', 'Length', 'Count'])
    for hand, count in counts.iteritems():
        csvwriter.writerow([';'.join(hand), len(hand), count])                    

######################################################################

if __name__ == '__main__':
    all_attested_hands_to_csv('hands.csv')
    stats()    

